# Copyright 2025 Open Navigation LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Module for a description of a ComposableLifecycleNode."""

from typing import List
from typing import Optional

import launch
from launch.substitution import Substitution
from launch.utilities import perform_substitutions
from launch_ros.parameters_type import Parameters
from launch_ros.remap_rule_type import RemapRules
from launch_ros.utilities import LifecycleEventManager

from .composable_node import ComposableNode


class ComposableLifecycleNode(ComposableNode):
    """Describes a lifecycle node that can be loaded into a container with other nodes."""

    def __init__(
        self, *,
        autostart: bool = False,
        **kwargs
    ) -> None:
        """
        Initialize a ComposableLifecycleNode description.

        :param autostart: Whether to autostart lifecycle node in the activated state
        """
        super().__init__(**kwargs)

        self.__autostart = autostart
        self.__lifecycle_event_manager = None
        self.__node_name = super().node_name

    def init_lifecycle_event_manager(self, context: launch.LaunchContext) -> None:
        # LifecycleEventManager needs a pre-substitution node name
        self.__node_name = perform_substitutions(context, self.node_name)
        self.__lifecycle_event_manager = LifecycleEventManager(self)
        self.__lifecycle_event_manager.setup_lifecycle_manager(context)

    @property
    def package(self) -> List[Substitution]:
        """Get node package name as a sequence of substitutions to be performed."""
        return super().package

    @property
    def node_plugin(self) -> List[Substitution]:
        """Get node plugin name as a sequence of substitutions to be performed."""
        return super().node_plugin

    @property
    def node_name(self) -> Optional[List[Substitution]]:
        """Get node name as a sequence of substitutions to be performed."""
        return self.__node_name

    @property
    def node_namespace(self) -> Optional[List[Substitution]]:
        """Get node namespace as a sequence of substitutions to be performed."""
        return super().node_namespace

    @property
    def node_autostart(self):
        """Getter for autostart."""
        return self.__autostart

    @property
    def parameters(self) -> Optional[Parameters]:
        """Get node parameter YAML files or dicts with substitutions to be performed."""
        return super().parameters

    @property
    def remappings(self) -> Optional[RemapRules]:
        """Get node remapping rules as (from, to) tuples with substitutions to be performed."""
        return super().remappings

    @property
    def extra_arguments(self) -> Optional[Parameters]:
        """Get container extra arguments YAML files or dicts with substitutions to be performed."""
        return super().extra_arguments



# Copyright 2022 Open Source Robotics Foundation, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Module for the Parameter substitution."""

import collections.abc
from typing import Any
from typing import Iterable
from typing import List
from typing import Optional
from typing import Sequence
from typing import Text
from typing import Union

from launch.frontend import expose_substitution
from launch.launch_context import LaunchContext
from launch.some_substitutions_type import SomeSubstitutionsType
from launch.substitution import Substitution
from launch.substitutions.substitution_failure import SubstitutionFailure
from launch.utilities import normalize_to_list_of_substitutions
from launch.utilities import perform_substitutions


@expose_substitution('param')
class Parameter(Substitution):
    """
    Substitution that tries to get a parameter that was set by SetParameter.

    :raise: SubstitutionFailure when param is not found
    """

    def __init__(
        self,
        name: SomeSubstitutionsType,
        *,
        default: Optional[Union[Any, Iterable[Any]]] = None
    ) -> None:
        """Create a Parameter substitution."""
        super().__init__()
        self.__name = normalize_to_list_of_substitutions(name)
        if default is None:
            self.__default = default
        else:
            # convert any items in default that are not a Substitution or str to a str
            str_normalized_default = []  # type: List[Union[Text, Substitution]]
            definitely_iterable_default = ((),)  # type: Iterable[Any]
            if isinstance(default, collections.abc.Iterable):
                definitely_iterable_default = default
            else:
                definitely_iterable_default = (default,)
            for item in definitely_iterable_default:
                if isinstance(item, (str, Substitution)):
                    str_normalized_default.append(item)
                else:
                    str_normalized_default.append(str(item))
            # use normalize_to_list_of_substitutions to convert str to TextSubstitution's too
            self.__default = \
                normalize_to_list_of_substitutions(str_normalized_default)

    @classmethod
    def parse(cls, data: Sequence[SomeSubstitutionsType]):
        """Parse a Parameter substitution."""
        if len(data) < 1 or len(data) > 2:
            raise TypeError('param substitution expects 1 or 2 arguments')
        kwargs = {}
        kwargs['name'] = data[0]
        if len(data) == 2:
            kwargs['default'] = data[1]
        return cls, kwargs

    @property
    def name(self) -> List[Substitution]:
        """Getter for name."""
        return self.__name

    def describe(self) -> Text:
        """Return a description of this substitution as a string."""
        name_str = ' + '.join([sub.describe() for sub in self.name])
        return '{}(name={})'.format(self.__class__.__name__, name_str)

    def perform(self, context: LaunchContext) -> Text:
        """Perform the substitution."""
        name = perform_substitutions(context, self.name)
        params_container = context.launch_configurations.get('global_params', [])
        for param in params_container:
            if isinstance(param, tuple):
                if param[0] == name:
                    return str(param[1])

        if self.__default is None:
            raise SubstitutionFailure("parameter '{}' not found".format(name))
        else:
            return perform_substitutions(context, self.__default)



# Copyright 2024 Open Navigation LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Module for the management of LifecycleNode events."""

import functools
import threading
from typing import cast

import launch
import launch.logging

import lifecycle_msgs.msg
import lifecycle_msgs.srv

from ..events.lifecycle import ChangeState
from ..events.lifecycle import StateTransition

from ..ros_adapters import get_ros_node


class LifecycleEventManager:

    def __init__(self, lifecycle_node) -> None:
        """
        Construct a LifecycleEventManager utility.

        This utility emits some event(s) in certain circumstances:

        - :class:`launch.events.lifecycle.StateTransition`:

            - this event is emitted when a message is published to the
              "/<name>/transition_event" topic, indicating the lifecycle
              node represented by this action changed state

        This utility also handles some events related to lifecycle:

        - :class:`launch.events.lifecycle.ChangeState`

          - this event can be targeted to a single lifecycle node, or more than
            one, or even all lifecycle nodes, and it requests the targeted nodes
            to change state, see its documentation for more details.

        :param lifecycle_node: The lifecycle node to event manage.
        """
        self.__logger = launch.logging.get_logger(__name__)
        self.__lifecycle_node = lifecycle_node
        self.__rclpy_subscription = None
        self.__rclpy_change_state_client = None

    @property
    def node_name(self):
        return self.__lifecycle_node.node_name

    def _on_transition_event(self, context, msg):
        try:
            event = StateTransition(action=self.__lifecycle_node, msg=msg)
            context.asyncio_loop.call_soon_threadsafe(lambda: context.emit_event_sync(event))
        except Exception as exc:
            self.__logger.error(
                "Exception in handling of 'lifecycle.msg.TransitionEvent': {}".format(exc))

    def _call_change_state(self, request, context: launch.LaunchContext):
        while not self.__rclpy_change_state_client.wait_for_service(timeout_sec=1.0):
            if context.is_shutdown:
                self.__logger.warning(
                    "Abandoning wait for the '{}' service, due to shutdown.".format(
                        self.__rclpy_change_state_client.srv_name),
                )
                return

        # Asynchronously wait so that we can periodically check for shutdown.
        event = threading.Event()

        def unblock(future):
            nonlocal event
            event.set()

        response_future = self.__rclpy_change_state_client.call_async(request)
        response_future.add_done_callback(unblock)

        while not event.wait(1.0):
            if context.is_shutdown:
                self.__logger.warning(
                    "Abandoning wait for the '{}' service response, due to shutdown.".format(
                        self.__rclpy_change_state_client.srv_name),
                )
                response_future.cancel()
                return

        if response_future.exception() is not None:
            raise response_future.exception()
        response = response_future.result()

        if not response.success:
            self.__logger.error(
                "Failed to make transition '{}' for LifecycleNode '{}'".format(
                    ChangeState.valid_transitions[request.transition.id],
                    self.node_name,
                )
            )

    def _on_change_state_event(self, context: launch.LaunchContext) -> None:
        typed_event = cast(ChangeState, context.locals.event)
        if not typed_event.lifecycle_node_matcher(self.__lifecycle_node):
            return None
        request = lifecycle_msgs.srv.ChangeState.Request()
        request.transition.id = typed_event.transition_id
        context.add_completion_future(
            context.asyncio_loop.run_in_executor(None, self._call_change_state, request, context))

    def setup_lifecycle_manager(self, context: launch.LaunchContext) -> None:
        node = get_ros_node(context)

        # Create a subscription to monitor the state changes of the subprocess.
        self.__rclpy_subscription = node.create_subscription(
            lifecycle_msgs.msg.TransitionEvent,
            '{}/transition_event'.format(self.node_name),
            functools.partial(self._on_transition_event, context),
            10)

        # Create a service client to change state on demand.
        self.__rclpy_change_state_client = node.create_client(
            lifecycle_msgs.srv.ChangeState,
            '{}/change_state'.format(self.node_name))

        # Register an event handler to change states on a ChangeState lifecycle event.
        context.register_event_handler(launch.EventHandler(
            matcher=lambda event: isinstance(event, ChangeState),
            entities=[launch.actions.OpaqueFunction(function=self._on_change_state_event)],
        ))



# Copyright 2018 Open Source Robotics Foundation, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Module for the LifecycleNode action."""

from typing import List
from typing import Optional

import launch
from launch import SomeSubstitutionsType
from launch.action import Action
import launch.logging

import lifecycle_msgs.msg
import lifecycle_msgs.srv

from .lifecycle_transition import LifecycleTransition
from .node import Node

from ..utilities import LifecycleEventManager


class LifecycleNode(Node):
    """Action that executes a ROS lifecycle node."""

    def __init__(
        self,
        *,
        name: SomeSubstitutionsType,
        namespace: SomeSubstitutionsType,
        autostart: bool = False,
        **kwargs
    ) -> None:
        """
        Construct a LifecycleNode action.

        Almost all of the arguments are passed to :class:`Node` and eventually
        to :class:`launch.actions.ExecuteProcess`, so see the documentation of
        those classes for additional details.

        This action additionally emits some event(s) in certain circumstances:

        - :class:`launch.events.lifecycle.StateTransition`:

            - this event is emitted when a message is published to the
              "/<name>/transition_event" topic, indicating the lifecycle
              node represented by this action changed state

        This action also handles some events related to lifecycle:

        - :class:`launch.events.lifecycle.ChangeState`

          - this event can be targeted to a single lifecycle node, or more than
            one, or even all lifecycle nodes, and it requests the targeted nodes
            to change state, see its documentation for more details.

        :param name: The name of the lifecycle node.
          Although it defaults to None it is a required parameter and the default will be removed
          in a future release.
        :param namespace: The ROS namespace for this Node
        :param autostart: Whether or not to automatically transition to the 'active' state.
        """
        super().__init__(name=name, namespace=namespace, **kwargs)
        self.__logger = launch.logging.get_logger(__name__)
        self.__autostart = autostart
        self.__lifecycle_event_manager = None

    @property
    def node_autostart(self):
        """Getter for autostart."""
        return self.__autostart

    @property
    def is_lifecycle_node(self):
        return True

    def get_lifecycle_event_manager(self):
        return self.__lifecycle_event_manager

    def execute(self, context: launch.LaunchContext) -> Optional[List[Action]]:
        """
        Execute the action.

        Delegated to :meth:`launch.actions.ExecuteProcess.execute`.
        """
        self._perform_substitutions(context)  # ensure self.node_name is expanded
        if '<node_name_unspecified>' in self.node_name:
            raise RuntimeError('node_name unexpectedly incomplete for lifecycle node')

        self.__lifecycle_event_manager = LifecycleEventManager(self)
        self.__lifecycle_event_manager.setup_lifecycle_manager(context)

        # If autostart is enabled, transition to the 'active' state.
        autostart_actions = None
        if self.node_autostart:
            autostart_actions = [
                LifecycleTransition(
                    lifecycle_node_names=[self.node_name],
                    transition_ids=[lifecycle_msgs.msg.Transition.TRANSITION_CONFIGURE,
                                    lifecycle_msgs.msg.Transition.TRANSITION_ACTIVATE]
                ),
            ]

        # Delegate execution to Node and ExecuteProcess.
        node_actions = super().execute(context)  # type: Optional[List[Action]]
        if node_actions is not None and autostart_actions is not None:
            return node_actions + autostart_actions
        if node_actions is not None:
            return node_actions
        if autostart_actions is not None:
            return autostart_actions
        return None



# Copyright 2022 Open Source Robotics Foundation, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Module for the `SetROSLogDir` action."""

import os
from typing import List

from launch import Action
from launch import Substitution
from launch.frontend import Entity
from launch.frontend import expose_action
from launch.frontend import Parser
from launch.launch_context import LaunchContext
from launch.some_substitutions_type import SomeSubstitutionsType
from launch.utilities import normalize_to_list_of_substitutions
from launch.utilities import perform_substitutions
from rclpy.logging import get_logging_directory


@expose_action('set_ros_log_dir')
class SetROSLogDir(Action):
    """
    Action that sets the ros log directory.

    This is done by setting the ROS_LOG_DIR environment variable, which will
    influence all processes started after that time, in which ros was initialized.

    This can be used in combination with launch.actions.GroupAction and its
    ``scoped=true`` option to provide scoped changes to this environment
    variable.

    Note this will not affect nodes loaded into component containers which were
    started before this action is executed.
    """

    def __init__(
        self,
        new_log_dir: SomeSubstitutionsType,
        **kwargs
    ) -> None:
        """
        Create a SetROSLogDir action.

        :param new_log_dir: new log directory, if absolute it sets the ros log dir,
            but if it is a relative path, then it is joined with the current ros log dir.
        """
        super().__init__(**kwargs)
        self.__log_dir = normalize_to_list_of_substitutions(new_log_dir)

    @classmethod
    def parse(cls, entity: Entity, parser: Parser):
        """Return `SetROSLogDir` action and kwargs for constructing it."""
        _, kwargs = super().parse(entity, parser)
        kwargs['new_log_dir'] = parser.parse_substitution(entity.get_attr('new_log_dir'))
        return cls, kwargs

    @property
    def log_dir(self) -> List[Substitution]:
        """Getter for self.__log_dir."""
        return self.__log_dir

    def execute(self, context: LaunchContext):
        """Execute the action."""
        new_log_dir = perform_substitutions(context, self.log_dir)
        current_rclpy_logging_directory = get_logging_directory()
        # Prefer ROS_LOG_DIR over what rclpy reports, but fall back to that if not set.
        current_log_dir = os.environ.get('ROS_LOG_DIR', current_rclpy_logging_directory)
        # If new_log_dir is abs, then current_log_dir will be truncated and not used.
        log_dir = os.path.join(current_log_dir, new_log_dir)
        assert os.path.isabs(log_dir)
        os.environ['ROS_LOG_DIR'] = log_dir



# Copyright 2022 Open Source Robotics Foundation, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import functools

from typing import Iterable
from typing import List
from typing import Optional
from typing import Union


import launch
from launch import LaunchContext, SomeSubstitutionsType
from launch.action import Action
from launch.actions import EmitEvent, RegisterEventHandler
from launch.utilities import normalize_to_list_of_substitutions
from launch.utilities import perform_substitutions

from launch_ros.event_handlers import OnStateTransition
from launch_ros.events.lifecycle import ChangeState, StateTransition
from launch_ros.events.matchers import matches_node_name
from lifecycle_msgs.msg import Transition


class LifecycleTransition(Action):
    """An action that simplifies execution of lifecyle transitions."""

    transition_targets = {
        Transition.TRANSITION_CONFIGURE:
            {'start_state': 'configuring', 'goal_state': 'inactive'},
        Transition.TRANSITION_CLEANUP:
            {'start_state': 'cleaningup', 'goal_state': 'unconfigured'},
        Transition.TRANSITION_ACTIVATE:
            {'start_state': 'activating', 'goal_state': 'active'},
        Transition.TRANSITION_DEACTIVATE:
            {'start_state': 'deactivating', 'goal_state': 'inactive'},
        Transition.TRANSITION_UNCONFIGURED_SHUTDOWN:
            {'start_state': 'shuttingdown', 'goal_state': 'finalized'},
        Transition.TRANSITION_INACTIVE_SHUTDOWN:
            {'start_state': 'shuttingdown', 'goal_state': 'finalized'},
        Transition.TRANSITION_ACTIVE_SHUTDOWN:
            {'start_state': 'shuttingdown', 'goal_state': 'finalized'},
    }

    def __init__(
        self,
        *,
        lifecycle_node_names: Iterable[SomeSubstitutionsType],
        transition_ids: Iterable[Union[int, SomeSubstitutionsType]],
        **kwargs
    ) -> None:
        """
        Construct a LifecycleTransition action.

        The action will execute the passed in lifecycle transition for the
        lifecycle nodes with the indicated node names. The action will emit
        an event that triggers the first lifecycle transition of each node
        wait that the node reaches the transition goal and trigger the next
        transition in the list.
        You need to make sure, that the sequence of lifecyle transition you
        pass in is possible.

        :param lifecycle_node_names: The names of the lifecycle nodes to transition
        :param transitions_ids: The transitions to be executed.
        """
        super().__init__(**kwargs)
        if len(transition_ids) == 0:
            raise ValueError('No transition_ids provided.')

        if len(lifecycle_node_names) == 0:
            raise ValueError('No lifecycle_node_names provided.')

        self.__lifecycle_node_names = [
            normalize_to_list_of_substitutions(name)
            for name in lifecycle_node_names]
        transition_ids = [
            str(transition_id)
            if isinstance(transition_id, int) else transition_id
            for transition_id in transition_ids]
        self.__transition_ids = [
            normalize_to_list_of_substitutions(transition_id)
            for transition_id in transition_ids]

        self.__event_handlers = {}
        self.__logger = launch.logging.get_logger(__name__)

    def _remove_event_handlers(
            self,
            context: LaunchContext,
            node_name: str,
            reason: str = None):
        """Remove all consequent transitions if error occurs."""
        if reason is not None:
            self.__logger.info(
                f"Stopping transitions for {node_name} because '{reason}'")

        for event_handler in self.__event_handlers[node_name]:
            # Unregister event handlers and ignore failures, as these are
            # already unregistered event handlers.
            try:
                context.unregister_event_handler(event_handler=event_handler)
            except ValueError:
                pass

    def execute(
        self,
        context: launch.LaunchContext
    ) -> Optional[List[Action]]:
        """
        Execute the LifecycleTransition action.

        :return Returns a list of actions to be executed to achieve specified transitions.
          These are EventHandlers and EventEmitters for ChangeState and
          StateTransition events of the nodes indicated.
        """
        lifecycle_node_names = [
            perform_substitutions(context, name)
            for name in self.__lifecycle_node_names]
        subs_transition_ids = [
            perform_substitutions(context, id_)
            for id_ in self.__transition_ids]
        transition_ids = []
        for tid in subs_transition_ids:
            try:
                transition_ids.append(int(tid))
            except ValueError:
                raise ValueError(
                    f'expected integer for lifecycle transition, got {tid}')

        emit_actions = {}
        actions: List[Action] = []

        # Create EmitEvents for ChangeStates and store
        for name in lifecycle_node_names:
            own_emit_actions = []
            for tid in transition_ids:
                change_event = ChangeState(
                    lifecycle_node_matcher=matches_node_name(name),
                    transition_id=tid)
                emit_action = EmitEvent(
                    event=change_event
                )
                own_emit_actions.append(emit_action)
            emit_actions[name] = own_emit_actions
            self.__event_handlers[name] = []
        # Create Transition EventHandlers and Registration actions
        i = 1
        for tid in transition_ids:
            # Create Transition handler for all indicated nodes
            for node_name in lifecycle_node_names:

                states = self.transition_targets[tid]
                event_handler = None
                # For all transitions except the last, emit next ChangeState Event
                if i < len(transition_ids):
                    event_handler = OnStateTransition(
                        matcher=match_node_name_start_goal(
                            node_name,
                            states['start_state'],
                            states['goal_state']),
                        entities=[
                            emit_actions[node_name][i]],
                        handle_once=True
                    )
                # For last transition emit Log message and remove untriggered error handlers
                else:
                    event_handler = OnStateTransition(
                        matcher=match_node_name_start_goal(
                            node_name,
                            states['start_state'],
                            states['goal_state']),
                        entities=[
                            launch.actions.OpaqueFunction(
                                function=functools.partial(
                                    self._remove_event_handlers,
                                    node_name=node_name
                                )
                            ),
                        ],
                        handle_once=True
                    )
                self.__event_handlers[node_name].append(event_handler)
                # Create register event handler action
                register_action = RegisterEventHandler(
                    event_handler=event_handler)
                # Append to actions
                actions.append(register_action)
            # increment next ChangeState action by one
            i += 1
        # Create Error processing event handlers.
        for node_name in lifecycle_node_names:
            event_handler = \
                launch.EventHandler(
                    matcher=match_node_name_goal(node_name, 'errorprocessing'),
                    entities=[
                        launch.actions.OpaqueFunction(
                            function=functools.partial(
                                self._remove_event_handlers,
                                node_name=node_name,
                                reason='error occured during transitions'
                            )
                        )
                    ],
                    handle_once=True,
                )
            self.__event_handlers[node_name].append(event_handler)
            context.register_event_handler(event_handler=event_handler)

        # Add first Emit actions to actions
        for node_name in lifecycle_node_names:
            actions.append(emit_actions[node_name][0])

        return actions


def match_node_name_start_goal(node_name: str, start_state: str, goal_state: str):
    if not node_name.startswith('/'):
        node_name = f'/{node_name}'
    return lambda event: (
        isinstance(event, StateTransition) and
        (event.action.node_name == node_name) and
        (event.goal_state == goal_state) and
        (event.start_state == start_state)
    )


def match_node_name_goal(node_name: str, goal_state: str):
    if not node_name.startswith('/'):
        node_name = f'/{node_name}'
    return lambda event: (
        isinstance(event, StateTransition) and
        (event.action.node_name == node_name) and
        (event.goal_state == goal_state)
    )
